#include "ff.h"  /* Declarations of FatFs API */
#include "sdcard.h"  /* Declarations of disk I/O functions */
#include <stdarg.h>

/* File access control feature */
#if _FS_LOCK
typedef struct {
  FATFS *fs;  /* Object ID 1, volume (NULL:blank entry) */
  uLong clu;  /* Object ID 2, directory */
  uShort idx;  /* Object ID 3, directory index */
  uShort ctr;  /* Object open counter, 0:none, 0x01..0xFF:read mode open count, 0x100:write mode */
} FILESEM;
#endif

/* DBCS code ranges and SBCS extend character conversion table */
//  Installed CODE PAGE is 1252, Latin 1 (Windows) (in ff.cpp)
//      other code pages in file ff.c http://elm-chan.org/fsw/ff/ff10a.zip
#define _DF1S  0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0xAd,0x9B,0x8C,0x9D,0xAE,0x9F, \
  0xA0,0x21,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
  0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
  0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0x9F}

/* Character code support macros */
#define IsUpper(c)  (((c)>='A')&&((c)<='Z'))
#define IsLower(c)  (((c)>='a')&&((c)<='z'))
#define IsDigit(c)  (((c)>='0')&&((c)<='9'))

#if _DF1S  /* Code page is DBCS */

#ifdef _DF2S  /* Two 1st byte areas */
#define IsDBCS1(c)  (((uChar)(c) >= _DF1S && (uChar)(c) <= _DF1E) || ((uChar)(c) >= _DF2S && (uChar)(c) <= _DF2E))
#else  /* One 1st byte area */
#define IsDBCS1(c)  ((uChar)(c) >= _DF1S && (uChar)(c) <= _DF1E)
#endif

#ifdef _DS3S  /* Three 2nd byte areas */
#define IsDBCS2(c)  (((uChar)(c) >= _DS1S && (uChar)(c) <= _DS1E) || ((uChar)(c) >= _DS2S && (uChar)(c) <= _DS2E) || ((uChar)(c) >= _DS3S && (uChar)(c) <= _DS3E))
#else  /* Two 2nd byte areas */
#define IsDBCS2(c)  (((uChar)(c) >= _DS1S && (uChar)(c) <= _DS1E) || ((uChar)(c) >= _DS2S && (uChar)(c) <= _DS2E))
#endif

#else  /* Code page is SBCS */

#define IsDBCS1(c)  0
#define IsDBCS2(c)  0

#endif /* _DF1S */

/* Name status flags */
#define NS  11  /* Index of name status byte in fn[] */
#define NS_LOSS  0x01  /* Out of 8.3 format */
#define NS_LFN  0x02  /* Force to create LFN entry */
#define NS_LAST  0x04  /* Last segment */
#define NS_BODY  0x08  /* Lower case flag (body) */
#define NS_EXT  0x10  /* Lower case flag (ext) */
#define NS_DOT  0x20  /* Dot entry */

/* FAT sub-type boundaries */
#define MIN_FAT16  4086U  /* Minimum number of clusters for FAT16 */
#define  MIN_FAT32  65526U  /* Minimum number of clusters for FAT32 */

/* FatFs refers the members in the FAT structures as byte array instead of
 / structure member because the structure is not binary compatible between
 / different platforms */

#define BS_jmpBoot  0  /* Jump instruction (3) */
#define BS_OEMName  3  /* OEM name (8) */
#define BPB_BytsPerSec  11  /* Sector size [byte] (2) */
#define BPB_SecPerClus  13  /* Cluster size [sector] (1) */
#define BPB_RsvdSecCnt  14  /* Size of reserved area [sector] (2) */
#define BPB_NumFATs  16  /* Number of FAT copies (1) */
#define BPB_RootEntCnt  17  /* Number of root directory entries for FAT12/16 (2) */
#define BPB_TotSec16  19  /* Volume size [sector] (2) */
#define BPB_Media  21  /* Media descriptor (1) */
#define BPB_FATSz16  22  /* FAT size [sector] (2) */
#define BPB_SecPerTrk  24  /* Track size [sector] (2) */
#define BPB_NumHeads  26  /* Number of heads (2) */
#define BPB_HiddSec  28  /* Number of special hidden sectors (4) */
#define BPB_TotSec32  32  /* Volume size [sector] (4) */
#define BS_DrvNum  36  /* Physical drive number (2) */
#define BS_BootSig  38  /* Extended boot signature (1) */
#define BS_VolID  39  /* Volume serial number (4) */
#define BS_VolLab  43  /* Volume label (8) */
#define BS_FilSysType  54  /* File system type (1) */
#define BPB_FATSz32  36  /* FAT size [sector] (4) */
#define BPB_ExtFlags  40  /* Extended flags (2) */
#define BPB_FSVer  42  /* File system version (2) */
#define BPB_RootClus  44  /* Root directory first cluster (4) */
#define BPB_FSInfo  48  /* Offset of FSINFO sector (2) */
#define BPB_BkBootSec  50  /* Offset of backup boot sector (2) */
#define BS_DrvNum32  64  /* Physical drive number (2) */
#define BS_BootSig32  66  /* Extended boot signature (1) */
#define BS_VolID32  67  /* Volume serial number (4) */
#define BS_VolLab32  71  /* Volume label (8) */
#define BS_FilSysType32  82  /* File system type (1) */
#define  FSI_LeadSig  0  /* FSI: Leading signature (4) */
#define  FSI_StrucSig  484  /* FSI: Structure signature (4) */
#define  FSI_Free_Count  488  /* FSI: Number of free clusters (4) */
#define  FSI_Nxt_Free  492  /* FSI: Last allocated cluster (4) */
#define MBR_Table  446  /* MBR: Partition table offset (2) */
#define  SZ_PTE  16  /* MBR: Size of a partition table entry */
#define BS_55AA  510  /* Boot sector signature (2) */

#define  DIR_Name  0  /* Short file name (11) */
#define  DIR_Attr  11  /* Attribute (1) */
#define  DIR_NTres  12  /* NT flag (1) */
#define DIR_CrtTimeTenth  13  /* Created time sub-second (1) */
#define  DIR_CrtTime  14  /* Created time (2) */
#define  DIR_CrtDate  16  /* Created date (2) */
#define DIR_LstAccDate  18  /* Last accessed date (2) */
#define  DIR_FstClusHI  20  /* Higher 16-bit of first cluster (2) */
#define  DIR_WrtTime  22  /* Modified time (2) */
#define  DIR_WrtDate  24  /* Modified date (2) */
#define  DIR_FstClusLO  26  /* Lower 16-bit of first cluster (2) */
#define  DIR_FileSize  28  /* File size (4) */
#define  LDIR_Ord  0  /* LFN entry order and LLE flag (1) */
#define  LDIR_Attr  11  /* LFN attribute (1) */
#define  LDIR_Type  12  /* LFN type (1) */
#define  LDIR_Chksum  13  /* Sum of corresponding SFN entry */
#define  LDIR_FstClusLO  26  /* Filled by zero (0) */
#define  SZ_DIR  32  /* Size of a directory entry */
#define  LLE  0x40  /* Last long entry flag in LDIR_Ord */
#define  DDE  0xE5  /* Deleted directory entry mark in DIR_Name[0] */
#define  NDDE  0x05  /* Replacement of the character collides with DDE */

/*------------------------------------------------------------*/
/* Module private work area  */
/*------------------------------------------------------------*/

SdSpi _SdSpi;

static
FATFS *FatFs[1];  /* Pointer to the file system objects (logical drives) */
static
uShort Fsid;  /* File system mount ID */

#if _FS_LOCK
static
FILESEM  Files[_FS_LOCK];  /* Open object lock semaphores */
#endif

#define  DEF_NAMEBUF  uChar sfn[12]
#define INIT_BUF(dobj)  (dobj).fn = sfn
#define  FREE_BUF()

#ifdef _EXCVT
static
const uChar ExCvt[] = _EXCVT;  /* Upper conversion table for extended characters */
#endif

/*-----------------------------------------------------------------------*/
/* String functions  */
/*-----------------------------------------------------------------------*/

/* Copy memory to memory */
static
void mem_cpy (void* dst, const void* src, uInt cnt) {
  uChar *d = (uChar*)dst;
  const uChar *s = (const uChar*)src;

  while (cnt--)
    *d++ = *s++;
}

/* Fill memory */
static
void mem_set (void* dst, int val, uInt cnt) {
  uChar *d = (uChar*)dst;

  while (cnt--)
    *d++ = (uChar)val;
}

/* Compare memory to memory */
static
int mem_cmp (const void* dst, const void* src, uInt cnt) {
  const uChar *d = (const uChar *)dst, *s = (const uChar *)src;
  int r = 0;

  while (cnt-- && (r = *d++ - *s++) == 0) ;
  return r;
}

/* Check if chr is contained in the string */
static
int chk_chr (const char* str, int chr) {
  while (*str && *str != chr) str++;
  return *str;
}

/*-----------------------------------------------------------------------*/
/* File lock control functions  */
/*-----------------------------------------------------------------------*/
#if _FS_LOCK

static
FRESULT chk_lock (  /* Check if the file can be accessed */
DIR* dp,  /* Directory object pointing the file to be checked */
int acc  /* Desired access (0:Read, 1:Write, 2:Delete/Rename) */
)
{
  uInt i, be;

  /* Search file semaphore table */
  for (i = be = 0; i < _FS_LOCK; i++) {
    if (Files[i].fs) {  /* Existing entry */
      if (Files[i].fs == dp->fs &&  /* Check if the object matched with an open object */
      Files[i].clu == dp->sclust &&
        Files[i].idx == dp->index) break;
    } 
    else {  /* Blank entry */
      be = 1;
    }
  }
  if (i == _FS_LOCK)  /* The object is not opened */
    return (be || acc == 2) ? FR_OK : FR_TOO_MANY_OPEN_FILES;  /* Is there a blank entry for new object? */

  /* The object has been opened. Reject any open against writing file and all write mode open */
  return (acc || Files[i].ctr == 0x100) ? FR_LOCKED : FR_OK;
}

static
int enq_lock (void)  /* Check if an entry is available for a new object */
{
  uInt i;

  for (i = 0; i < _FS_LOCK && Files[i].fs; i++) ;
  return (i == _FS_LOCK) ? 0 : 1;
}

static
uInt inc_lock (  /* Increment object open counter and returns its index (0:Internal error) */
DIR* dp,  /* Directory object pointing the file to register or increment */
int acc  /* Desired access (0:Read, 1:Write, 2:Delete/Rename) */
)
{
  uInt i;

  for (i = 0; i < _FS_LOCK; i++) {  /* Find the object */
    if (Files[i].fs == dp->fs &&
      Files[i].clu == dp->sclust &&
      Files[i].idx == dp->index) break;
  }

  if (i == _FS_LOCK) {  /* Not opened. Register it as new. */
    for (i = 0; i < _FS_LOCK && Files[i].fs; i++) ;
    if (i == _FS_LOCK) return 0;  /* No free entry to register (int err) */
    Files[i].fs = dp->fs;
    Files[i].clu = dp->sclust;
    Files[i].idx = dp->index;
    Files[i].ctr = 0;
  }

  if (acc && Files[i].ctr) return 0;  /* Access violation (int err) */

  Files[i].ctr = acc ? 0x100 : Files[i].ctr + 1;  /* Set semaphore value */

  return i + 1;
}

static
FRESULT dec_lock (  /* Decrement object open counter */
uInt i  /* Semaphore index (1..) */
)
{
  uShort n;
  FRESULT res;

  if (--i < _FS_LOCK) {  /* Shift index number origin from 0 */
    n = Files[i].ctr;
    if (n == 0x100) n = 0;  /* If write mode open, delete the entry */
    if (n) n--;  /* Decrement read mode open count */
    Files[i].ctr = n;
    if (!n) Files[i].fs = 0;  /* Delete the entry if open count gets zero */
    res = FR_OK;
  } 
  else {
    res = FR_INT_ERR;  /* Invalid index nunber */
  }
  return res;
}

static
void clear_lock (  /* Clear lock entries of the volume */
FATFS *fs
)
{
  uInt i;

  for (i = 0; i < _FS_LOCK; i++) {
    if (Files[i].fs == fs) Files[i].fs = 0;
  }
}
#endif

/*-----------------------------------------------------------------------*/
/* Move/Flush disk access window in the file system object  */
/*-----------------------------------------------------------------------*/
static
FRESULT sync_window (
FATFS* fs  /* File system object */
)
{
  uLong wsect;
  uInt nf;
  if (fs->wflag) {  /* Write back the sector if it is dirty */
    wsect = fs->winsect;  /* Current sector number */
    if (_SdSpi.disk_write(fs->win, wsect, 1))
      return FR_DISK_ERR;
    fs->wflag = 0;
    if (wsect - fs->fatbase < fs->fsize) {  /* Is it in the FAT area? */
      for (nf = fs->n_fats; nf >= 2; nf--) {  /* Reflect the change to all FAT copies */
        wsect += fs->fsize;
        _SdSpi.disk_write(fs->win, wsect, 1);
      }
    }
  }
  return FR_OK;
}

static
FRESULT move_window (
FATFS* fs,  /* File system object */
uLong sector  /* Sector number to make appearance in the fs->win[] */
)
{
  if (sector != fs->winsect) {  /* Changed current window */
    if (sync_window(fs) != FR_OK)
      return FR_DISK_ERR;
    if (_SdSpi.disk_read(fs->win, sector, 1))
      return FR_DISK_ERR;
    fs->winsect = sector;
  }

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Synchronize file system and strage device  */
/*-----------------------------------------------------------------------*/
static
FRESULT sync_fs (  /* FR_OK: successful, FR_DISK_ERR: failed */
FATFS* fs  /* File system object */
)
{
  FRESULT res;

  res = sync_window(fs);
  if (res == FR_OK) {
    /* Update FSINFO sector if needed */
    if (fs->fs_type == FS_FAT32 && fs->fsi_flag == 1) {
      /* Create FSINFO structure */
      mem_set(fs->win, 0, (uInt)512);
      ST_uShort(fs->win+BS_55AA, 0xAA55);
      ST_uLong(fs->win+FSI_LeadSig, 0x41615252);
      ST_uLong(fs->win+FSI_StrucSig, 0x61417272);
      ST_uLong(fs->win+FSI_Free_Count, fs->free_clust);
      ST_uLong(fs->win+FSI_Nxt_Free, fs->last_clust);
      /* Write it into the FSINFO sector */
      fs->winsect = fs->volbase + 1;
      _SdSpi.disk_write(fs->win, fs->winsect, 1);
      fs->fsi_flag = 0;
    }
    /* Make sure that no pending write process in the physical drive */
    if (_SdSpi.disk_ioctl() != RES_OK)
      res = FR_DISK_ERR;
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* Get sector# from cluster#  */
/*-----------------------------------------------------------------------*/
uLong clust2sect (  /* !=0: Sector number, 0: Failed - invalid cluster# */
FATFS* fs,  /* File system object */
uLong clst  /* Cluster# to be converted */
)
{
  clst -= 2;
  if (clst >= (fs->n_fatent - 2)) return 0;  /* Invalid cluster# */
  return clst * fs->csize + fs->database;
}

/*-----------------------------------------------------------------------*/
/* FAT access - Read value of a FAT entry  */
/*-----------------------------------------------------------------------*/
uLong get_fat (  /* 0xFFFFFFFF:Disk error, 1:Internal error, Else:Cluster status */
FATFS* fs,  /* File system object */
uLong clst  /* Cluster# to get the link information */
)
{
  uInt wc, bc;
  uChar *p;

  if (clst < 2 || clst >= fs->n_fatent)  /* Check range */
    return 1;

  switch (fs->fs_type) {
  case FS_FAT12 :
    bc = (uInt)clst; 
    bc += bc / 2;
    if (move_window(fs, fs->fatbase + (bc / (uInt)512))) break;
    wc = fs->win[bc % (uInt)512]; 
    bc++;
    if (move_window(fs, fs->fatbase + (bc / (uInt)512))) break;
    wc |= fs->win[bc % (uInt)512] << 8;
    return clst & 1 ? wc >> 4 : (wc & 0xFFF);

  case FS_FAT16 :
    if (move_window(fs, fs->fatbase + (clst / ((uInt)512 / 2)))) break;
    p = &fs->win[clst * 2 % (uInt)512];
    return LD_uShort(p);

  case FS_FAT32 :
    if (move_window(fs, fs->fatbase + (clst / ((uInt)512 / 4)))) break;
    p = &fs->win[clst * 4 % (uInt)512];
    return LD_uLong(p) & 0x0FFFFFFF;

  default:
    return 1;
  }

  return 0xFFFFFFFF;  /* An error occurred at the disk I/O layer */
}

/*-----------------------------------------------------------------------*/
/* FAT access - Change value of a FAT entry  */
/*-----------------------------------------------------------------------*/
FRESULT put_fat (
FATFS* fs,  /* File system object */
uLong clst,  /* Cluster# to be changed in range of 2 to fs->n_fatent - 1 */
uLong val  /* New value to mark the cluster */
)
{
  uInt bc;
  uChar *p;
  FRESULT res;

  if (clst < 2 || clst >= fs->n_fatent) {  /* Check range */
    res = FR_INT_ERR;

  } 
  else {
    switch (fs->fs_type) {
    case FS_FAT12 :
      bc = (uInt)clst; 
      bc += bc / 2;
      res = move_window(fs, fs->fatbase + (bc / (uInt)512));
      if (res != FR_OK) break;
      p = &fs->win[bc % (uInt)512];
      *p = (clst & 1) ? ((*p & 0x0F) | ((uChar)val << 4)) : (uChar)val;
      bc++;
      fs->wflag = 1;
      res = move_window(fs, fs->fatbase + (bc / (uInt)512));
      if (res != FR_OK) break;
      p = &fs->win[bc % (uInt)512];
      *p = (clst & 1) ? (uChar)(val >> 4) : ((*p & 0xF0) | ((uChar)(val >> 8) & 0x0F));
      break;

    case FS_FAT16 :
      res = move_window(fs, fs->fatbase + (clst / ((uInt)512 / 2)));
      if (res != FR_OK) break;
      p = &fs->win[clst * 2 % (uInt)512];
      ST_uShort(p, (uShort)val);
      break;

    case FS_FAT32 :
      res = move_window(fs, fs->fatbase + (clst / ((uInt)512 / 4)));
      if (res != FR_OK) break;
      p = &fs->win[clst * 4 % (uInt)512];
      val |= LD_uLong(p) & 0xF0000000;
      ST_uLong(p, val);
      break;

    default :
      res = FR_INT_ERR;
    }
    fs->wflag = 1;
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* FAT handling - Remove a cluster chain  */
/*-----------------------------------------------------------------------*/
static
FRESULT remove_chain (
FATFS* fs,  /* File system object */
uLong clst  /* Cluster# to remove a chain from */
)
{
  FRESULT res;
  uLong nxt;

  if (clst < 2 || clst >= fs->n_fatent) {  /* Check range */
    res = FR_INT_ERR;

  } 
  else {
    res = FR_OK;
    while (clst < fs->n_fatent) {  /* Not a last link? */
      nxt = get_fat(fs, clst);  /* Get cluster status */
      if (nxt == 0) break;  /* Empty cluster? */
      if (nxt == 1) { 
        res = FR_INT_ERR; 
        break; 
      }  /* Internal error? */
      if (nxt == 0xFFFFFFFF) { 
        res = FR_DISK_ERR; 
        break; 
      }  /* Disk error? */
      res = put_fat(fs, clst, 0);  /* Mark the cluster "empty" */
      if (res != FR_OK) break;
      if (fs->free_clust != 0xFFFFFFFF) {  /* Update FSINFO */
        fs->free_clust++;
        fs->fsi_flag |= 1;
      }
      clst = nxt;  /* Next cluster */
    }
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* FAT handling - Stretch or Create a cluster chain  */
/*-----------------------------------------------------------------------*/
static
uLong create_chain (  /* 0:No free cluster, 1:Internal error, 0xFFFFFFFF:Disk error, >=2:New cluster# */
FATFS* fs,  /* File system object */
uLong clst  /* Cluster# to stretch. 0 means create a new chain. */
)
{
  uLong cs, ncl, scl;
  FRESULT res;

  if (clst == 0) {  /* Create a new chain */
    scl = fs->last_clust;  /* Get suggested start point */
    if (!scl || scl >= fs->n_fatent) scl = 1;
  }
  else {  /* Stretch the current chain */
    cs = get_fat(fs, clst);  /* Check the cluster status */
    if (cs < 2) return 1;  /* Invalid value */
    if (cs == 0xFFFFFFFF) return cs;  /* A disk error occurred */
    if (cs < fs->n_fatent) return cs;  /* It is already followed by next cluster */
    scl = clst;
  }

  ncl = scl;  /* Start cluster */
  for (;;) {
    ncl++;  /* Next cluster */
    if (ncl >= fs->n_fatent) {  /* Check wrap around */
      ncl = 2;
      if (ncl > scl) return 0;  /* No free cluster */
    }
    cs = get_fat(fs, ncl);  /* Get the cluster status */
    if (cs == 0) break;  /* Found a free cluster */
    if (cs == 0xFFFFFFFF || cs == 1)/* An error occurred */
      return cs;
    if (ncl == scl) return 0;  /* No free cluster */
  }

  res = put_fat(fs, ncl, 0x0FFFFFFF);  /* Mark the new cluster "last link" */
  if (res == FR_OK && clst != 0) {
    res = put_fat(fs, clst, ncl);  /* Link it to the previous one if needed */
  }
  if (res == FR_OK) {
    fs->last_clust = ncl;  /* Update FSINFO */
    if (fs->free_clust != 0xFFFFFFFF) {
      fs->free_clust--;
      fs->fsi_flag |= 1;
    }
  } 
  else {
    ncl = (res == FR_DISK_ERR) ? 0xFFFFFFFF : 1;
  }

  return ncl;  /* Return new cluster number or error code */
}

/*-----------------------------------------------------------------------*/
/* Directory handling - Set directory index  */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_sdi (
DIR* dp,  /* Pointer to directory object */
uInt idx  /* Index of directory table */
)
{
  uLong clst, sect;
  uInt ic;

  dp->index = (uShort)idx;  /* Current index */
  clst = dp->sclust;  /* Table start cluster (0:root) */
  if (clst == 1 || clst >= dp->fs->n_fatent)  /* Check start cluster range */
    return FR_INT_ERR;
  if (!clst && dp->fs->fs_type == FS_FAT32)  /* Replace cluster# 0 with root cluster# if in FAT32 */
    clst = dp->fs->dirbase;

  if (clst == 0) {  /* Static table (root-directory in FAT12/16) */
    if (idx >= dp->fs->n_rootdir)  /* Is index out of range? */
      return FR_INT_ERR;
    sect = dp->fs->dirbase;
  }
  else {  /* Dynamic table (root-directory in FAT32 or sub-directory) */
    ic = (uInt)512 / SZ_DIR * dp->fs->csize;  /* Entries per cluster */
    while (idx >= ic) {  /* Follow cluster chain */
      clst = get_fat(dp->fs, clst);  /* Get next cluster */
      if (clst == 0xFFFFFFFF) return FR_DISK_ERR;  /* Disk error */
      if (clst < 2 || clst >= dp->fs->n_fatent)  /* Reached to end of table or internal error */
        return FR_INT_ERR;
      idx -= ic;
    }
    sect = clust2sect(dp->fs, clst);
  }
  dp->clust = clst;  /* Current cluster# */
  if (!sect) return FR_INT_ERR;
  dp->sect = sect + idx / ((uInt)512 / SZ_DIR);  /* Sector# of the directory entry */
  dp->dir = dp->fs->win + (idx % ((uInt)512 / SZ_DIR)) * SZ_DIR;  /* Ptr to the entry in the sector */

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Directory handling - Move directory table index next  */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_next (  /* FR_OK:Succeeded, FR_NO_FILE:End of table, FR_DENIED:Could not stretch */
DIR* dp,  /* Pointer to the directory object */
int stretch  /* 0: Do not stretch table, 1: Stretch table if needed */
)
{
  uLong clst;
  uInt i;

  i = dp->index + 1;
  if (!(i & 0xFFFF) || !dp->sect)  /* Report EOT when index has reached 65535 */
    return FR_NO_FILE;

  if (!(i % ((uInt)512 / SZ_DIR))) {  /* Sector changed? */
    dp->sect++;  /* Next sector */

    if (!dp->clust) {  /* Static table */
      if (i >= dp->fs->n_rootdir)  /* Report EOT if it reached end of static table */
        return FR_NO_FILE;
    }
    else {  /* Dynamic table */
      if (((i / ((uInt)512 / SZ_DIR)) & (dp->fs->csize - 1)) == 0) {  /* Cluster changed? */
        clst = get_fat(dp->fs, dp->clust);  /* Get next cluster */
        if (clst <= 1) return FR_INT_ERR;
        if (clst == 0xFFFFFFFF) return FR_DISK_ERR;
        if (clst >= dp->fs->n_fatent) {  /* If it reached end of dynamic table, */
          uInt c;
          if (!stretch) return FR_NO_FILE;  /* If do not stretch, report EOT */
          clst = create_chain(dp->fs, dp->clust);  /* Stretch cluster chain */
          if (clst == 0) return FR_DENIED;  /* No free cluster */
          if (clst == 1) return FR_INT_ERR;
          if (clst == 0xFFFFFFFF) return FR_DISK_ERR;
          /* Clean-up stretched table */
          if (sync_window(dp->fs)) return FR_DISK_ERR;/* Flush disk access window */
          mem_set(dp->fs->win, 0, (uInt)512);  /* Clear window buffer */
          dp->fs->winsect = clust2sect(dp->fs, clst);  /* Cluster start sector */
          for (c = 0; c < dp->fs->csize; c++) {  /* Fill the new cluster with 0 */
            dp->fs->wflag = 1;
            if (sync_window(dp->fs)) return FR_DISK_ERR;
            dp->fs->winsect++;
          }
          dp->fs->winsect -= c;  /* Rewind window offset */
        }
        dp->clust = clst;  /* Initialize data for new cluster */
        dp->sect = clust2sect(dp->fs, clst);
      }
    }
  }

  dp->index = (uShort)i;  /* Current index */
  dp->dir = dp->fs->win + (i % ((uInt)512 / SZ_DIR)) * SZ_DIR;  /* Current entry in the window */

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Directory handling - Reserve directory entry  */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_alloc (
DIR* dp,  /* Pointer to the directory object */
uInt nent  /* Number of contiguous entries to allocate (1-21) */
)
{
  FRESULT res;
  uInt n;


  res = dir_sdi(dp, 0);
  if (res == FR_OK) {
    n = 0;
    do {
      res = move_window(dp->fs, dp->sect);
      if (res != FR_OK) break;
      if (dp->dir[0] == DDE || dp->dir[0] == 0) {  /* Is it a blank entry? */
        if (++n == nent) break;  /* A block of contiguous entries is found */
      } 
      else {
        n = 0;  /* Not a blank entry. Restart to search */
      }
      res = dir_next(dp, 1);  /* Next entry with table stretch enabled */
    } 
    while (res == FR_OK);
  }
  if (res == FR_NO_FILE) res = FR_DENIED;  /* No directory entry to allocate */
  return res;
}

/*-----------------------------------------------------------------------*/
/* Directory handling - Load/Store start cluster number  */
/*-----------------------------------------------------------------------*/

static
uLong ld_clust (
FATFS* fs,  /* Pointer to the fs object */
uChar* dir  /* Pointer to the directory entry */
)
{
  uLong cl;

  cl = LD_uShort(dir+DIR_FstClusLO);
  if (fs->fs_type == FS_FAT32)
    cl |= (uLong)LD_uShort(dir+DIR_FstClusHI) << 16;

  return cl;
}

static
void st_clust (
uChar* dir,  /* Pointer to the directory entry */
uLong cl  /* Value to be set */
)
{
  ST_uShort(dir+DIR_FstClusLO, cl);
  ST_uShort(dir+DIR_FstClusHI, cl >> 16);
}

/*-----------------------------------------------------------------------*/
/* Directory handling - Find an object in the directory  */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_find (
DIR* dp  /* Pointer to the directory object linked to the file name */
)
{
  FRESULT res;
  uChar c, *dir;

  res = dir_sdi(dp, 0);  /* Rewind directory object */
  if (res != FR_OK) return res;

  do {
    res = move_window(dp->fs, dp->sect);
    if (res != FR_OK) break;
    dir = dp->dir;  /* Ptr to the directory entry of current index */
    c = dir[DIR_Name];
    if (c == 0) { 
      res = FR_NO_FILE; 
      break; 
    }  /* Reached to end of table */
    if (!(dir[DIR_Attr] & AM_VOL) && !mem_cmp(dir, dp->fn, 11)) /* Is it a valid entry? */
      break;
    res = dir_next(dp, 0);  /* Next entry */
  } 
  while (res == FR_OK);

  return res;
}

/*-----------------------------------------------------------------------*/
/* Register an object to the directory  */
/*-----------------------------------------------------------------------*/
static
FRESULT dir_register (  /* FR_OK:Successful, FR_DENIED:No free entry or too many SFN collision, FR_DISK_ERR:Disk error */
DIR* dp  /* Target directory with object name to be created */
)
{
  FRESULT res;
  res = dir_alloc(dp, 1);  /* Allocate an entry for SFN */

  if (res == FR_OK) {  /* Set SFN entry */
    res = move_window(dp->fs, dp->sect);
    if (res == FR_OK) {
      mem_set(dp->dir, 0, SZ_DIR);  /* Clean the entry */
      mem_cpy(dp->dir, dp->fn, 11);  /* Put SFN */
      dp->fs->wflag = 1;
    }
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* Pick a segment and create the object name in directory form  */
/*-----------------------------------------------------------------------*/

static
FRESULT create_name (
DIR* dp,  /* Pointer to the directory object */
const TCHAR** path  /* Pointer to pointer to the segment in the path string */
)
{
  uChar b, c, d, *sfn;
  uInt ni, si, i;
  const char *p;

  /* Create file name in directory form */
  for (p = *path; *p == '/' || *p == '\\'; p++) ;  /* Strip duplicated separator */
  sfn = dp->fn;
  mem_set(sfn, ' ', 11);
  si = i = b = 0; 
  ni = 8;
  for (;;) {
    c = (uChar)p[si++];
    if (c <= ' ' || c == '/' || c == '\\') break;  /* Break on end of segment */
    if (c == '.' || i >= ni) {
      if (ni != 8 || c != '.') return FR_INVALID_NAME;
      i = 8; 
      ni = 11;
      b <<= 2; 
      continue;
    }
    if (c >= 0x80) {  /* Extended character? */
      b |= 3;  /* Eliminate NT flag */
#ifdef _EXCVT
      c = ExCvt[c - 0x80];  /* To upper extended characters (SBCS cfg) */
#else
#if !_DF1S
      return FR_INVALID_NAME;  /* Reject extended characters (ASCII cfg) */
#endif
#endif
    }
    if (IsDBCS1(c)) {  /* Check if it is a DBC 1st byte (always false on SBCS cfg) */
      d = (uChar)p[si++];  /* Get 2nd byte */
      if (!IsDBCS2(d) || i >= ni - 1)  /* Reject invalid DBC */
        return FR_INVALID_NAME;
      sfn[i++] = c;
      sfn[i++] = d;
    } 
    else {  /* Single byte code */
      if (chk_chr("\"*+,:;<=>\?[]|\x7F", c))  /* Reject illegal chrs for SFN */
        return FR_INVALID_NAME;
      if (IsUpper(c)) {  /* ASCII large capital? */
        b |= 2;
      } 
      else {
        if (IsLower(c)) {  /* ASCII small capital? */
          b |= 1; 
          c -= 0x20;
        }
      }
      sfn[i++] = c;
    }
  }
  *path = &p[si];  /* Return pointer to the next segment */
  c = (c <= ' ') ? NS_LAST : 0;  /* Set last segment flag if end of path */

  if (!i) return FR_INVALID_NAME;  /* Reject nul string */
  if (sfn[0] == DDE) sfn[0] = NDDE;  /* When first character collides with DDE, replace it with 0x05 */

  if (ni == 8) b <<= 2;
  if ((b & 0x03) == 0x01) c |= NS_EXT;  /* NT flag (Name extension has only small capital) */
  if ((b & 0x0C) == 0x04) c |= NS_BODY;  /* NT flag (Name body has only small capital) */

  sfn[NS] = c;  /* Store NT flag, File name is created */

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Follow a file path  */
/*-----------------------------------------------------------------------*/

static
FRESULT follow_path (  /* FR_OK(0): successful, !=0: error code */
DIR* dp,  /* Directory object to return last directory and found object */
const TCHAR* path  /* Full-path string to find a file or directory */
)
{
  FRESULT res;
  uChar *dir, ns;

  if (*path == '/' || *path == '\\')  /* Strip heading separator if exist */
    path++;
  dp->sclust = 0;  /* Always start from the root directory */

  if ((uInt)*path < ' ') {  /* Null path name is the origin directory itself */
    res = dir_sdi(dp, 0);
    dp->dir = 0;
  } 
  else {  /* Follow path */
    for (;;) {
      res = create_name(dp, &path);  /* Get a segment name of the path */
      if (res != FR_OK) break;
      res = dir_find(dp);  /* Find an object with the sagment name */
      ns = dp->fn[NS];
      if (res != FR_OK) {  /* Failed to find the object */
        if (res == FR_NO_FILE) {  /* Object is not found */
          if (ns & NS_DOT) {  /* If dot entry is not exist, */
            dp->sclust = 0; 
            dp->dir = 0;  /* it is the root directory and stay there */
            if (!(ns & NS_LAST)) continue;  /* Continue to follow if not last segment */
            res = FR_OK;  /* Ended at the root directroy. Function completed. */
          } 
          else {  /* Could not find the object */
            if (!(ns & NS_LAST)) res = FR_NO_PATH;  /* Adjust error code if not last segment */
          }
        }
        break;
      }
      if (ns & NS_LAST) break;  /* Last segment matched. Function completed. */
      dir = dp->dir;  /* Follow the sub-directory */
      if (!(dir[DIR_Attr] & AM_DIR)) {  /* It is not a sub-directory and cannot follow */
        res = FR_NO_PATH; 
        break;
      }
      dp->sclust = ld_clust(dp->fs, dir);
    }
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* Get logical drive number from path name  */
/*-----------------------------------------------------------------------*/
static
int get_ldnumber (  /* Returns logical drive number (-1:invalid drive) */
const TCHAR** path  /* Pointer to pointer to the path name */
)
{
  const TCHAR *tp, *tt;
  uInt i;
  int vol = -1;

  if (*path) {  /* If the pointer is not a null */
    for (tt = *path; (uInt)*tt >= '!' && *tt != ':'; tt++) ;  // for (tt = *path; (uInt)*tt >= (_USE_LFN ? ' ' : '!') && *tt != ':'; tt++) ;  /* Find ':' in the path */
    if (*tt == ':') {  /* If a ':' is exist in the path name */
      tp = *path;
      i = *tp++ - '0'; 
      if (i < 10 && tp == tt) {  /* Is there a numeric drive id? */
        if (i < 1) {  /* If a drive id is found, get the value and strip it */
          vol = (int)i;
          *path = ++tt;
        }
      } 
      return vol;
    }
    vol = 0;  /* Drive 0 */
  }
  return vol;
}

/*-----------------------------------------------------------------------*/
/* Load a sector and check if it is an FAT boot sector  */
/*-----------------------------------------------------------------------*/
static
uChar check_fs (  /* 0:FAT boor sector, 1:Valid boor sector but not FAT, 2:Not a boot sector, 3:Disk error */
FATFS* fs,  /* File system object */
uLong sect  /* Sector# (lba) to check if it is an FAT boot record or not */
)
{
  fs->wflag = 0; 
  fs->winsect = 0xFFFFFFFF;  /* Invaidate window */
  if (move_window(fs, sect) != FR_OK)  /* Load boot record */
    return 3;

  if (LD_uShort(&fs->win[BS_55AA]) != 0xAA55)  /* Check boot record signature (always placed at offset 510 even if the sector size is >512) */
    return 2;

  if ((LD_uLong(&fs->win[BS_FilSysType]) & 0xFFFFFF) == 0x544146)  /* Check "FAT" string */
    return 0;
  if ((LD_uLong(&fs->win[BS_FilSysType32]) & 0xFFFFFF) == 0x544146)  /* Check "FAT" string */
    return 0;

  return 1;
}

/*-----------------------------------------------------------------------*/
/* Find logical drive and check if the volume is mounted  */
/*-----------------------------------------------------------------------*/
static
FRESULT find_volume (  /* FR_OK(0): successful, !=0: any error occurred */
FATFS** rfs,  /* Pointer to pointer to the found file system object */
const TCHAR** path,  /* Pointer to pointer to the path name (drive number) */
uChar wmode  /* !=0: Check write protection for write access */
)
{
  uChar fmt;
  int vol;
  DSTATUS stat;
  uLong bsect, fasize, tsect, sysect, nclst, szbfat;
  uShort nrsv;
  FATFS *fs;

  /* Get logical drive number from the path name */
  *rfs = 0;
  vol = get_ldnumber(path);
  if (vol < 0) return FR_INVALID_DRIVE;

  /* Check if the file system object is valid or not */
  fs = FatFs[vol];  /* Get pointer to the file system object */
  if (!fs) return FR_NOT_ENABLED;  /* Is the file system object available? */

  *rfs = fs;  /* Return pointer to the file system object */

  if (fs->fs_type) {  /* If the volume has been mounted */
    stat = _SdSpi.disk_status();
    if (!(stat & STA_NOINIT)) {  /* and the physical drive is kept initialized */
      if (wmode && (stat & STA_PROTECT))  /* Check write protection if needed */
        return FR_WRITE_PROTECTED;
      return FR_OK;  /* The file system object is valid */
    }
  }

  /* The file system object is not valid. */
  /* Following code attempts to mount the volume. (analyze BPB and initialize the fs object) */

  fs->fs_type = 0;  /* Clear the file system object */
  fs->drv = LD2PD(vol);  /* Bind the logical drive and a physical drive */
  stat = _SdSpi.disk_initialize();  /* Initialize the physical drive */
  if (stat & STA_NOINIT)  /* Check if the initialization succeeded */
    return FR_NOT_READY;  /* Failed to initialize due to no medium or hard error */
  if (wmode && (stat & STA_PROTECT))  /* Check disk write protection if needed */
    return FR_WRITE_PROTECTED;
  /* Find an FAT partition on the drive. Supports only generic partitioning, FDISK and SFD. */
  bsect = 0;
  fmt = check_fs(fs, bsect);  /* Load sector 0 and check if it is an FAT boot sector as SFD */
  if (fmt == 1 || (!fmt && (LD2PT(vol)))) {  /* Not an FAT boot sector or forced partition number */
    uInt i;
    uLong br[4];

    for (i = 0; i < 4; i++) {  /* Get partition offset */
      uChar *pt = fs->win+MBR_Table + i * SZ_PTE;
      br[i] = pt[4] ? LD_uLong(&pt[8]) : 0;
    }
    i = LD2PT(vol);  /* Partition number: 0:auto, 1-4:forced */
    if (i) i--;
    do {  /* Find an FAT volume */
      bsect = br[i];
      fmt = bsect ? check_fs(fs, bsect) : 2;  /* Check the partition */
    } 
    while (!LD2PT(vol) && fmt && ++i < 4);
  }
  if (fmt == 3) return FR_DISK_ERR;  /* An error occured in the disk I/O layer */
  if (fmt) return FR_NO_FILESYSTEM;  /* No FAT volume is found */

  /* An FAT volume is found. Following code initializes the file system object */

  if (LD_uShort(fs->win+BPB_BytsPerSec) != (uInt)512)  /* (BPB_BytsPerSec must be equal to the physical sector size) */
    return FR_NO_FILESYSTEM;

  fasize = LD_uShort(fs->win+BPB_FATSz16);  /* Number of sectors per FAT */
  if (!fasize) fasize = LD_uLong(fs->win+BPB_FATSz32);
  fs->fsize = fasize;

  fs->n_fats = fs->win[BPB_NumFATs];  /* Number of FAT copies */
  if (fs->n_fats != 1 && fs->n_fats != 2)  /* (Must be 1 or 2) */
    return FR_NO_FILESYSTEM;
  fasize *= fs->n_fats;  /* Number of sectors for FAT area */

  fs->csize = fs->win[BPB_SecPerClus];  /* Number of sectors per cluster */
  if (!fs->csize || (fs->csize & (fs->csize - 1)))  /* (Must be power of 2) */
    return FR_NO_FILESYSTEM;

  fs->n_rootdir = LD_uShort(fs->win+BPB_RootEntCnt);  /* Number of root directory entries */
  if (fs->n_rootdir % ((uInt)512 / SZ_DIR))  /* (Must be sector aligned) */
    return FR_NO_FILESYSTEM;

  tsect = LD_uShort(fs->win+BPB_TotSec16);  /* Number of sectors on the volume */
  if (!tsect) tsect = LD_uLong(fs->win+BPB_TotSec32);

  nrsv = LD_uShort(fs->win+BPB_RsvdSecCnt);  /* Number of reserved sectors */
  if (!nrsv) return FR_NO_FILESYSTEM;  /* (Must not be 0) */

  /* Determine the FAT sub type */
  sysect = nrsv + fasize + fs->n_rootdir / ((uInt)512 / SZ_DIR);  /* RSV+FAT+DIR */
  if (tsect < sysect) return FR_NO_FILESYSTEM;  /* (Invalid volume size) */
  nclst = (tsect - sysect) / fs->csize;  /* Number of clusters */
  if (!nclst) return FR_NO_FILESYSTEM;  /* (Invalid volume size) */
  fmt = FS_FAT12;
  if (nclst >= MIN_FAT16) fmt = FS_FAT16;
  if (nclst >= MIN_FAT32) fmt = FS_FAT32;

  /* Boundaries and Limits */
  fs->n_fatent = nclst + 2;  /* Number of FAT entries */
  fs->volbase = bsect;  /* Volume start sector */
  fs->fatbase = bsect + nrsv;  /* FAT start sector */
  fs->database = bsect + sysect;  /* Data start sector */
  if (fmt == FS_FAT32) {
    if (fs->n_rootdir) return FR_NO_FILESYSTEM;  /* (BPB_RootEntCnt must be 0) */
    fs->dirbase = LD_uLong(fs->win+BPB_RootClus);  /* Root directory start cluster */
    szbfat = fs->n_fatent * 4;  /* (Needed FAT size) */
  } 
  else {
    if (!fs->n_rootdir)  return FR_NO_FILESYSTEM;  /* (BPB_RootEntCnt must not be 0) */
    fs->dirbase = fs->fatbase + fasize;  /* Root directory start sector */
    szbfat = (fmt == FS_FAT16) ?  /* (Needed FAT size) */
    fs->n_fatent * 2 : fs->n_fatent * 3 / 2 + (fs->n_fatent & 1);
  }
  if (fs->fsize < (szbfat + ((uInt)512 - 1)) / (uInt)512)  /* (BPB_FATSz must not be less than needed) */
    return FR_NO_FILESYSTEM;

  /* Initialize cluster allocation information */
  fs->last_clust = fs->free_clust = 0xFFFFFFFF;

  /* Get fsinfo if available */
  fs->fsi_flag = 0x80;
#if (_FS_NOFSINFO & 3) != 3
  if (fmt == FS_FAT32  /* Enable FSINFO only if FAT32 and BPB_FSInfo is 1 */
  && LD_uShort(fs->win+BPB_FSInfo) == 1
    && move_window(fs, bsect + 1) == FR_OK)
  {
    fs->fsi_flag = 0;
    if (LD_uShort(fs->win+BS_55AA) == 0xAA55  /* Load FSINFO data if available */
    && LD_uLong(fs->win+FSI_LeadSig) == 0x41615252
      && LD_uLong(fs->win+FSI_StrucSig) == 0x61417272)
    {
#if (_FS_NOFSINFO & 1) == 0
      fs->free_clust = LD_uLong(fs->win+FSI_Free_Count);
#endif
#if (_FS_NOFSINFO & 2) == 0
      fs->last_clust = LD_uLong(fs->win+FSI_Nxt_Free);
#endif
    }
  }
#endif

  fs->fs_type = fmt;  /* FAT sub-type */
  fs->id = ++Fsid;  /* File system mount ID */
#if _FS_LOCK  /* Clear file lock semaphores */
  clear_lock(fs);
#endif

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Check if the file/directory object is valid or not  */
/*-----------------------------------------------------------------------*/
static
FRESULT validate (  /* FR_OK(0): The object is valid, !=0: Invalid */
void* obj  /* Pointer to the object FIL/DIR to check validity */
)
{
  FIL *fil = (FIL*)obj;  /* Assuming offset of .fs and .id in the FIL/DIR structure is identical */

  if (!fil || !fil->fs || !fil->fs->fs_type || fil->fs->id != fil->id)
    return FR_INVALID_OBJECT;

  if (_SdSpi.disk_status() & STA_NOINIT)
    return FR_NOT_READY;

  return FR_OK;
}

/*--------------------------------------------------------------------------
 Public Functions
 --------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Mount/Unmount a Logical Drive  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_mount (
FATFS* fs,  /* Pointer to the file system object (NULL:unmount)*/
const TCHAR* path,  /* Logical drive number to be mounted/unmounted */
uChar opt  /* 0:Do not mount (delayed mount), 1:Mount immediately */
)
{
  FATFS *cfs;
  int vol;
  FRESULT res;
  const TCHAR *rp = path;

  vol = get_ldnumber(&rp);
  if (vol < 0) return FR_INVALID_DRIVE;
  cfs = FatFs[vol];  /* Pointer to fs object */

  if (cfs) {
#if _FS_LOCK
    clear_lock(cfs);
#endif
    cfs->fs_type = 0;  /* Clear old fs object */
  }

  if (fs) {
    fs->fs_type = 0;  /* Clear new fs object */
  }
  FatFs[vol] = fs;  /* Register new fs object */

  if (!fs || opt != 1) return FR_OK;  /* Do not mount now, it will be mounted later */

  res = find_volume(&fs, &path, 0);  /* Force mounted the volume */
  return res;
}

/*-----------------------------------------------------------------------*/
/* Open or Create a File  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_open (
FIL* fp,  /* Pointer to the blank file object */
const TCHAR* path,  /* Pointer to the file name */
uChar mode  /* Access mode and file open mode flags */
)
{
  FRESULT res;
  DIR dj;
  uChar *dir;
  DEF_NAMEBUF;

  if (!fp) return FR_INVALID_OBJECT;
  fp->fs = 0;  /* Clear file object */

  /* Get logical drive number */
  mode &= FA_READ | FA_WRITE | FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW;
  res = find_volume(&dj.fs, &path, (uChar)(mode & ~FA_READ));
  if (res == FR_OK) {
    INIT_BUF(dj);
    res = follow_path(&dj, path);  /* Follow the file path */
    dir = dj.dir;
    /* R/W configuration */
    if (res == FR_OK) {
      if (!dir)  /* Default directory itself */
        res = FR_INVALID_NAME;
#if _FS_LOCK
      else
        res = chk_lock(&dj, (mode & ~FA_READ) ? 1 : 0);
#endif
    }
    /* Create or Open a file */
    if (mode & (FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW)) {
      uLong dw, cl;
      if (res != FR_OK) {  /* No file, create new */
        if (res == FR_NO_FILE)  /* There is no file to open, create a new entry */
#if _FS_LOCK
          res = enq_lock() ? dir_register(&dj) : FR_TOO_MANY_OPEN_FILES;
#else
        res = dir_register(&dj);
#endif
        mode |= FA_CREATE_ALWAYS;  /* File is created */
        dir = dj.dir;  /* New entry */
      }
      else {  /* Any object is already existing */
        if (dir[DIR_Attr] & (AM_RDO | AM_DIR))  /* Cannot overwrite it (R/O or DIR) */
          res = FR_DENIED;
        else {
          if (mode & FA_CREATE_NEW)  /* Cannot create as new file */
            res = FR_EXIST;
        }
      }
      if (res == FR_OK && (mode & FA_CREATE_ALWAYS)) {  /* Truncate it if overwrite mode */
        dw = get_fattime();  /* Created time */
        ST_uLong(dir+DIR_CrtTime, dw);
        dir[DIR_Attr] = 0;  /* Reset attribute */
        ST_uLong(dir+DIR_FileSize, 0);  /* size = 0 */
        cl = ld_clust(dj.fs, dir);  /* Get start cluster */
        st_clust(dir, 0);  /* cluster = 0 */
        dj.fs->wflag = 1;
        if (cl) {  /* Remove the cluster chain if exist */
          dw = dj.fs->winsect;
          res = remove_chain(dj.fs, cl);
          if (res == FR_OK) {
            dj.fs->last_clust = cl - 1;  /* Reuse the cluster hole */
            res = move_window(dj.fs, dw);
          }
        }
      }
    }
    else {  /* Open an existing file */
      if (res == FR_OK) {  /* Follow succeeded */
        if (dir[DIR_Attr] & AM_DIR) {  /* It is a directory */
          res = FR_NO_FILE;
        } 
        else {
          if ((mode & FA_WRITE) && (dir[DIR_Attr] & AM_RDO)) /* R/O violation */
            res = FR_DENIED;
        }
      }
    }
    if (res == FR_OK) {
      if (mode & FA_CREATE_ALWAYS)  /* Set file change flag if created or overwritten */
        mode |= FA__WRITTEN;
      fp->dir_sect = dj.fs->winsect;  /* Pointer to the directory entry */
      fp->dir_ptr = dir;
#if _FS_LOCK
      fp->lockid = inc_lock(&dj, (mode & ~FA_READ) ? 1 : 0);
      if (!fp->lockid) res = FR_INT_ERR;
#endif
    }

    FREE_BUF();

    if (res == FR_OK) {
      fp->flag = mode;  /* File access mode */
      fp->err = 0;  /* Clear error flag */
      fp->sclust = ld_clust(dj.fs, dir);  /* File start cluster */
      fp->fsize = LD_uLong(dir+DIR_FileSize);  /* File size */
      fp->fptr = 0;  /* File pointer */
      fp->dsect = 0;
      fp->fs = dj.fs;  /* Validate file object */
      fp->id = fp->fs->id;
    }
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* Read File  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_read (
FIL* fp,  /* Pointer to the file object */
void* buff,  /* Pointer to data buffer */
uInt btr,  /* Number of bytes to read */
uInt* br  /* Pointer to number of bytes read */
)
{
  FRESULT res;
  uLong clst, sect, remain;
  uInt rcnt, cc;
  uChar csect, *rbuff = (uChar*)buff;

  *br = 0;  /* Clear read byte counter */

  res = validate(fp);  /* Check validity */
  if (res != FR_OK) return res;
  if (fp->err)  /* Check error */
    return (FRESULT)fp->err;
  if (!(fp->flag & FA_READ))  /* Check access mode */
    return FR_DENIED;
  remain = fp->fsize - fp->fptr;
  if (btr > remain) btr = (uInt)remain;  /* Truncate btr by remaining bytes */

  for ( ;  btr;  /* Repeat until all data read */
  rbuff += rcnt, fp->fptr += rcnt, *br += rcnt, btr -= rcnt) {
      if ((fp->fptr % (uInt)512) == 0) {  /* On the sector boundary? */
      csect = (uChar)(fp->fptr / (uInt)512 & (fp->fs->csize - 1));  /* Sector offset in the cluster */
      if (!csect) {  /* On the cluster boundary? */
        if (fp->fptr == 0) {  /* On the top of the file? */
          clst = fp->sclust;  /* Follow from the origin */
        } 
        else {  /* Middle or end of the file */
          clst = get_fat(fp->fs, fp->clust);  /* Follow cluster chain on the FAT */
        }
        if (clst < 2)
        {
          fp->err = FR_INT_ERR;
          return res;
        }
        if (clst == 0xFFFFFFFF)
        {
          fp->err = FR_DISK_ERR;
          return res;
        }
        fp->clust = clst;  /* Update current cluster */
      }
      sect = clust2sect(fp->fs, fp->clust);  /* Get current sector */
      if (!sect)
      {
        fp->err = FR_INT_ERR;
        return res;
      }
      sect += csect;
      cc = btr / (uInt)512;  /* When remaining bytes >= sector size, */
      if (cc) {  /* Read maximum contiguous sectors directly */
        if (csect + cc > fp->fs->csize)  /* Clip at cluster boundary */
          cc = fp->fs->csize - csect;
        if (_SdSpi.disk_read(rbuff, sect, cc))
        {
          fp->err = FR_DISK_ERR;
          return res;
        }
        /* Replace one of the read sectors with cached data if it contains a dirty sector */
        if (fp->fs->wflag && fp->fs->winsect - sect < cc)
          mem_cpy(rbuff + ((fp->fs->winsect - sect) * (uInt)512), fp->fs->win, (uInt)512);
        rcnt = (uInt)512 * cc;  /* Number of bytes transferred */
        continue;
      }
      fp->dsect = sect;
    }
    rcnt = (uInt)512 - ((uInt)fp->fptr % (uInt)512);  /* Get partial sector data from sector buffer */
    if (rcnt > btr) rcnt = btr;
    if (move_window(fp->fs, fp->dsect))  /* Move sector window */
    {
      fp->err = FR_DISK_ERR;
      return res;
    }
    mem_cpy(rbuff, &fp->fs->win[fp->fptr % (uInt)512], rcnt);  /* Pick partial sector */
  }

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Write File  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_write (
FIL* fp,  /* Pointer to the file object */
const void *buff,  /* Pointer to the data to be written */
uInt btw,  /* Number of bytes to write */
uInt* bw  /* Pointer to number of bytes written */
)
{
  FRESULT res;
  uLong clst, sect;
  uInt wcnt, cc;
  const uChar *wbuff = (const uChar*)buff;
  uChar csect;

  *bw = 0;  /* Clear write byte counter */

  res = validate(fp);  /* Check validity */
  if (res != FR_OK) return res;
  if (fp->err)  /* Check error */
    return (FRESULT)fp->err;
  if (!(fp->flag & (FA_WRITE | FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW | FA_OPEN_EXISTING)))  /* Check access mode */
    return FR_DENIED;
  if (fp->fptr + btw < fp->fptr) btw = 0;  /* File size cannot reach 4GB */

  for ( ;  btw;  /* Repeat until all data written */
  wbuff += wcnt, fp->fptr += wcnt, *bw += wcnt, btw -= wcnt) {
      if ((fp->fptr % (uInt)512) == 0) {  /* On the sector boundary? */
      csect = (uChar)(fp->fptr / (uInt)512 & (fp->fs->csize - 1));  /* Sector offset in the cluster */
      if (!csect) {  /* On the cluster boundary? */
        if (fp->fptr == 0) {  /* On the top of the file? */
          clst = fp->sclust;  /* Follow from the origin */
          if (clst == 0)  /* When no cluster is allocated, */
            fp->sclust = clst = create_chain(fp->fs, 0);  /* Create a new cluster chain */
        } 
        else {  /* Middle or end of the file */
          clst = create_chain(fp->fs, fp->clust);  /* Follow or stretch cluster chain on the FAT */
        }
        if (clst == 0) break;  /* Could not allocate a new cluster (disk full) */
        if (clst == 1)
        {
          fp->err = FR_INT_ERR;
          return res;
        }

        if (clst == 0xFFFFFFFF)
        {
          fp->err = FR_DISK_ERR;
          return res;
        }

        fp->clust = clst;  /* Update current cluster */
      }
      if (fp->fs->winsect == fp->dsect && sync_window(fp->fs))  /* Write-back sector cache */
      {
        fp->err = FR_DISK_ERR;
        return res;
      }
      sect = clust2sect(fp->fs, fp->clust);  /* Get current sector */
      if (!sect)
      {
        fp->err = FR_INT_ERR;
        return res;
      }

      sect += csect;
      cc = btw / (uInt)512;  /* When remaining bytes >= sector size, */
      if (cc) {  /* Write maximum contiguous sectors directly */
        if (csect + cc > fp->fs->csize)  /* Clip at cluster boundary */
          cc = fp->fs->csize - csect;
        if (_SdSpi.disk_write(wbuff, sect, cc))
        {
          fp->err = FR_DISK_ERR;
          return res;
        }
        if (fp->fs->winsect - sect < cc) {  /* Refill sector cache if it gets invalidated by the direct write */
          mem_cpy(fp->fs->win, wbuff + ((fp->fs->winsect - sect) * (uInt)512), (uInt)512);
          fp->fs->wflag = 0;
        }
        wcnt = (uInt)512 * cc;  /* Number of bytes transferred */
        continue;
      }
      if (fp->fptr >= fp->fsize) {  /* Avoid silly cache filling at growing edge */
        if (sync_window(fp->fs))
        {
          fp->err = FR_DISK_ERR;
          return res;
        }

        fp->fs->winsect = sect;
      }
      fp->dsect = sect;
    }
    wcnt = (uInt)512 - ((uInt)fp->fptr % (uInt)512);/* Put partial sector into file I/O buffer */
    if (wcnt > btw) wcnt = btw;
    if (move_window(fp->fs, fp->dsect))  /* Move sector window */
    {
      fp->err = FR_DISK_ERR;
      return res;
    }
    mem_cpy(&fp->fs->win[fp->fptr % (uInt)512], wbuff, wcnt);  /* Fit partial sector */
    fp->fs->wflag = 1;
  }

  if (fp->fptr > fp->fsize) fp->fsize = fp->fptr;  /* Update file size if needed */
  fp->flag |= FA__WRITTEN;  /* Set file change flag */

  return FR_OK;
}

/*-----------------------------------------------------------------------*/
/* Synchronize the File  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_sync (
FIL* fp  /* Pointer to the file object */
)
{
  FRESULT res;
  uLong tm;
  uChar *dir;

  res = validate(fp);  /* Check validity of the object */
  if (res == FR_OK) {
    if (fp->flag & FA__WRITTEN) {  /* Has the file been written? */
      /* Write-back dirty buffer */
      /* Update the directory entry */
      res = move_window(fp->fs, fp->dir_sect);
      if (res == FR_OK) {
        dir = fp->dir_ptr;
        dir[DIR_Attr] |= AM_ARC;  /* Set archive bit */
        ST_uLong(dir+DIR_FileSize, fp->fsize);  /* Update file size */
        st_clust(dir, fp->sclust);  /* Update start cluster */
        tm = get_fattime();  /* Update updated time */
        ST_uLong(dir+DIR_WrtTime, tm);
        ST_uShort(dir+DIR_LstAccDate, 0);
        fp->flag &= ~FA__WRITTEN;
        fp->fs->wflag = 1;
        res = sync_fs(fp->fs);
      }
    }
  }

  return res;
}

/*-----------------------------------------------------------------------*/
/* Close File  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_close (
FIL *fp  /* Pointer to the file object to be closed */
)
{
  FRESULT res;

  res = f_sync(fp);  /* Flush cached data */
  if (res == FR_OK)
  {
    res = validate(fp);  /* Lock volume */
    if (res == FR_OK) {
#if _FS_LOCK
      res = dec_lock(fp->lockid);  /* Decrement file open counter */
      if (res == FR_OK)
#endif
        fp->fs = 0;  /* Invalidate file object */
    }
  }
  return res;
}

/*-----------------------------------------------------------------------*/
/* Seek File R/W Pointer  */
/*-----------------------------------------------------------------------*/
FRESULT SdFatFs::f_lseek (
FIL* fp,  /* Pointer to the file object */
uLong ofs  /* File pointer from top of file */
)
{
  FRESULT res;

  res = validate(fp);  /* Check validity of the object */
  if (res != FR_OK) return res;
  if (fp->err)  /* Check error */
    return (FRESULT)fp->err;

  /* Normal Seek */
  {
    uLong clst, bcs, nsect, ifptr;

    if (ofs > fp->fsize  /* In read-only mode, clip offset with the file size */
    && !(fp->flag & FA_WRITE)
      ) ofs = fp->fsize;

    ifptr = fp->fptr;
    fp->fptr = nsect = 0;
    if (ofs) {
      bcs = (uLong)fp->fs->csize * (uInt)512;  /* Cluster size (byte) */
      if (ifptr > 0 &&
        (ofs - 1) / bcs >= (ifptr - 1) / bcs) {  /* When seek to same or following cluster, */
        fp->fptr = (ifptr - 1) & ~(bcs - 1);  /* start from the current cluster */
        ofs -= fp->fptr;
        clst = fp->clust;
      } 
      else {  /* When seek to back cluster, */
        clst = fp->sclust;  /* start from the first cluster */
        if (clst == 0) {  /* If no cluster chain, create a new chain */
          clst = create_chain(fp->fs, 0);
          if (clst == 1)
          {
            fp->err = FR_INT_ERR;
            return res;
          }
          if (clst == 0xFFFFFFFF)
          {
            fp->err = FR_DISK_ERR;
            return res;
          }
          fp->sclust = clst;
        }
        fp->clust = clst;
      }
      if (clst != 0) {
        while (ofs > bcs) {  /* Cluster following loop */
          if (fp->flag & FA_WRITE) {  /* Check if in write mode or not */
            clst = create_chain(fp->fs, clst);  /* Force stretch if in write mode */
            if (clst == 0) {  /* When disk gets full, clip file size */
              ofs = bcs; 
              break;
            }
          } 
          else
            clst = get_fat(fp->fs, clst);  /* Follow cluster chain if not in write mode */
          if (clst == 0xFFFFFFFF)
          {
            fp->err = FR_DISK_ERR;
            return res;
          }
          if (clst <= 1 || clst >= fp->fs->n_fatent)
          {
            fp->err = FR_INT_ERR;
            return res;
          }
          fp->clust = clst;
          fp->fptr += bcs;
          ofs -= bcs;
        }
        fp->fptr += ofs;
        if (ofs % (uInt)512) {
          nsect = clust2sect(fp->fs, clst);  /* Current sector */
          if (!nsect)
          {
            fp->err = FR_DISK_ERR;
            return res;
          }
          nsect += ofs / (uInt)512;
        }
      }
    }
    if (fp->fptr % (uInt)512 && nsect != fp->dsect) {  /* Fill sector cache if needed */
      fp->dsect = nsect;
    }
    if (fp->fptr > fp->fsize) {  /* Set file change flag if the file size is extended */
      fp->fsize = fp->fptr;
      fp->flag |= FA__WRITTEN;
    }
  }

  return res;
}

#if _USE_STRFUNC
/*-----------------------------------------------------------------------*/
/* Get a string from the file  */
/*-----------------------------------------------------------------------*/
TCHAR* SdFatFs::f_gets (
TCHAR* buff,  /* Pointer to the string buffer to read */
int len,  /* Size of string buffer (characters) */
FIL* fp  /* Pointer to the file object */
)
{
  int n = 0;
  TCHAR c, *p = buff;
  uChar s[2];
  uInt rc;

  while (n < len - 1) {  /* Read characters until buffer gets filled */
    f_read(fp, s, 1, &rc);
    if (rc != 1) break;
    c = s[0];
    if (_USE_STRFUNC == 2 && c == '\r') continue;  /* Strip '\r' */
    *p++ = c;
    n++;
    if (c == '\n') break;  /* Break on EOL */
  }
  *p = 0;
  return n ? buff : 0;  /* When no data read (eof or error), return with error. */
}

/*-----------------------------------------------------------------------*/
/* Put a character to the file  */
/*-----------------------------------------------------------------------*/
struct putbuff{
  FIL* fp;
  int idx, nchr;
  uChar buf[64];
};

SdFatFs sdFat;
static
void putc_bfd (
putbuff* pb,
TCHAR c
)
{
  uInt bw;
  int i;

  if (_USE_STRFUNC == 2 && c == '\n')  /* LF -> CRLF conversion */
    putc_bfd(pb, '\r');

  i = pb->idx;  /* Buffer write index (-1:error) */
  if (i < 0) return;
  pb->buf[i++] = (uChar)c;
  if (i >= (int)(sizeof pb->buf) - 3) {  /* Write buffered characters to the file */
    sdFat.f_write(pb->fp, pb->buf, (uInt)i, &bw);
    i = (bw == (uInt)i) ? 0 : -1;
  }
  pb->idx = i;
  pb->nchr++;
}

int SdFatFs::f_putc (
TCHAR c,  /* A character to be output */
FIL* fp  /* Pointer to the file object */
)
{
  putbuff pb;
  uInt nw;

  pb.fp = fp;  /* Initialize output buffer */
  pb.nchr = pb.idx = 0;

  putc_bfd(&pb, c);  /* Put a character */

  if (  pb.idx >= 0  /* Flush buffered characters to the file */
  && f_write(pb.fp, pb.buf, (uInt)pb.idx, &nw) == FR_OK
    && (uInt)pb.idx == nw) return pb.nchr;
  return EOF;
}

/*-----------------------------------------------------------------------*/
/* Put a string to the file  */
/*-----------------------------------------------------------------------*/
int SdFatFs::f_puts (
const TCHAR* str,  /* Pointer to the string to be output */
FIL* fp  /* Pointer to the file object */
)
{
  putbuff pb;
  uInt nw;

  pb.fp = fp;  /* Initialize output buffer */
  pb.nchr = pb.idx = 0;

  while (*str)  /* Put the string */
    putc_bfd(&pb, *str++);

  if (  pb.idx >= 0  /* Flush buffered characters to the file */
  && f_write(pb.fp, pb.buf, (uInt)pb.idx, &nw) == FR_OK
    && (uInt)pb.idx == nw) return pb.nchr;
  return EOF;
}

/*-----------------------------------------------------------------------*/
/* Put a formatted string to the file  */
/*-----------------------------------------------------------------------*/
int SdFatFs::f_printf (
FIL* fp,  /* Pointer to the file object */
const TCHAR* fmt,  /* Pointer to the format string */
...  /* Optional arguments... */
)
{
  va_list arp;
  uChar f, r;
  uInt nw, i, j, w;
  uLong v;
  TCHAR c, d, s[16], *p;
  putbuff pb;

  pb.fp = fp;  /* Initialize output buffer */
  pb.nchr = pb.idx = 0;

  va_start(arp, fmt);

  for (;;) {
    c = *fmt++;
    if (c == 0) break;  /* End of string */
    if (c != '%') {  /* Non escape character */
      putc_bfd(&pb, c);
      continue;
    }
    w = f = 0;
    c = *fmt++;
    if (c == '0') {  /* Flag: '0' padding */
      f = 1; 
      c = *fmt++;
    } 
    else {
      if (c == '-') {  /* Flag: left justified */
        f = 2; 
        c = *fmt++;
      }
    }
    while (IsDigit(c)) {  /* Precision */
      w = w * 10 + c - '0';
      c = *fmt++;
    }
    if (c == 'l' || c == 'L') {  /* Prefix: Size is long int */
      f |= 4; 
      c = *fmt++;
    }
    if (!c) break;
    d = c;
    if (IsLower(d)) d -= 0x20;
    switch (d) {  /* Type is... */
    case 'S' :  /* String */
      p = va_arg(arp, TCHAR*);
      for (j = 0; p[j]; j++) ;
      if (!(f & 2)) {
        while (j++ < w) putc_bfd(&pb, ' ');
      }
      while (*p) putc_bfd(&pb, *p++);
      while (j++ < w) putc_bfd(&pb, ' ');
      continue;
    case 'C' :  /* Character */
      putc_bfd(&pb, (TCHAR)va_arg(arp, int)); 
      continue;
    case 'B' :  /* Binary */
      r = 2; 
      break;
    case 'O' :  /* Octal */
      r = 8; 
      break;
    case 'D' :  /* Signed decimal */
    case 'U' :  /* Unsigned decimal */
      r = 10; 
      break;
    case 'X' :  /* Hexdecimal */
      r = 16; 
      break;
    default:  /* Unknown type (pass-through) */
      putc_bfd(&pb, c); 
      continue;
    }

    /* Get an argument and put it in numeral */
    v = (f & 4) ? (uLong)va_arg(arp, long) : ((d == 'D') ? (uLong)(long)va_arg(arp, int) : (uLong)va_arg(arp, unsigned int));
    if (d == 'D' && (v & 0x80000000)) {
      v = 0 - v;
      f |= 8;
    }
    i = 0;
    do {
      d = (TCHAR)(v % r); 
      v /= r;
      if (d > 9) d += (c == 'x') ? 0x27 : 0x07;
      s[i++] = d + '0';
    } 
    while (v && i < sizeof s / sizeof s[0]);
    if (f & 8) s[i++] = '-';
    j = i; 
    d = (f & 1) ? '0' : ' ';
    while (!(f & 2) && j++ < w) putc_bfd(&pb, d);
    do putc_bfd(&pb, s[--i]); 
    while (i);
    while (j++ < w) putc_bfd(&pb, d);
  }

  va_end(arp);

  if (  pb.idx >= 0  /* Flush buffered characters to the file */
  && f_write(pb.fp, pb.buf, (uInt)pb.idx, &nw) == FR_OK
    && (uInt)pb.idx == nw) return pb.nchr;
  return EOF;
}

#endif /* _USE_STRFUNC */
