#include <SPI.h>
#include "sdcard.h"

uChar portSS = 10;  // SS SPI-PIN 
const uChar ClockDividerSlow = SPI_CLOCK_DIV64;
const uChar ClockDividerFast = SPI_CLOCK_DIV4;

volatile uChar media_status = STA_NOINIT;  // Disk status 
const uChar wait = 2;  // delay(time) - if init/read/write ERRORS increase the interval time

/*---------------------------
 Module Private Functions
 ---------------------------*/

// Definitions for MMC/SDC command 
#define CMD0  (0)  // GO_IDLE_STATE 
#define CMD1  (1)  // SEND_OP_COND (MMC) 
#define ACMD41  (0x80+41)  // SEND_OP_COND (SDC) 
#define CMD8  (8)  // SEND_IF_COND 
#define CMD9  (9)  // SEND_CSD 
#define CMD10  (10)  // SEND_CID 
#define CMD12  (12)  // STOP_TRANSMISSION 
#define ACMD13  (0x80+13)  // SD_STATUS (SDC) 
#define CMD16  (16)  // SET_BLOCKLEN 
#define CMD17  (17)  // READ_SINGLE_BLOCK 
#define CMD18  (18)  // READ_MULTIPLE_BLOCK 
#define CMD23  (23)  // SET_BLOCK_COUNT (MMC) 
#define ACMD23  (0x80+23)  // SET_WR_BLK_ERASE_COUNT (SDC) 
#define CMD24  (24)  // WRITE_BLOCK 
#define CMD25  (25)  // WRITE_MULTIPLE_BLOCK 
#define CMD55  (55)  // APP_CMD 
#define CMD58  (58)  // READ_OCR 

static uChar CardType;  // Card type flags 

/*----------------------
 Wait for card ready
 ----------------------*/
static int wait_ready (void)  // 1:OK, 0:Timeout 
{
  uInt tmr;

  for (tmr = 250; tmr; tmr--)
  {
    if (SPI.transfer(0xff) == 0xff) return 1;
    delay(wait);
  }
  return 0;
}

/*------------------------------------------
 Deselect the card and release SPI bus
 ------------------------------------------*/

static void deselect (void)
{
  digitalWrite(portSS, HIGH);
  SPI.transfer(0xFF);
}

/*---------------------------------------
 Select the card and wait for ready
 ---------------------------------------*/
static int select (void)  // 1:Successful, 0:Timeout 
{
  digitalWrite(portSS, LOW);
  SPI.transfer(0xFF); // Dummy clock (force DO enabled) 
  if (wait_ready()) return 1; // OK 
  deselect();
  return 0;  // Timeout 
}

/*---------------------------------
 Receive a data packet from MMC
 ----------------------------------*/
static int rcvr_datablock (uChar *buff, uInt btr)
{
  uChar d, tmr;
  for (tmr = 100; tmr; tmr--)  // Wait for data packet
  {
    d = SPI.transfer(0xFF);
    if (d != 0xff) break;
    delay(wait);
  } 
  if (d != 0xFE) return 0;  // If not valid data token, return with error 
  do
    *buff++ = SPI.transfer(0xff);
  while (--btr);
  SPI.transfer(0xFF);  // Discard CRC 
  SPI.transfer(0xFF);
  return 1;  // Return with success 
}

/*----------------------------
 Send a data packet to MMC
 ----------------------------*/
static int xmit_datablock (const uChar *buff, uChar token)
{
  uChar resp;
  if (!wait_ready()) return 0;
  SPI.transfer(token);  // Xmit data token 
  if (token != 0xFD)  // Is data token 
  {  
    uInt btr = 512;
    do
      SPI.transfer(*buff++);
    while (--btr);
    SPI.transfer(0xFF);  // CRC (Dummy) 
    SPI.transfer(0xFF);
    resp = SPI.transfer(0xFF);  // Reveive data response 
    if ((resp & 0x1F) != 0x05) return 0;  // If not accepted, return with error 
  }
  return 1;
}

/*-------------------------------
 Send a command packet to MMC
 -------------------------------*/
uChar send_cmd (uChar cmd, uLong arg)  // Returns R1 resp (bit7==1:Send failed) 
{
  uChar n, res;

  if (cmd & 0x80)  // ACMD<n> is the command sequense of CMD55-CMD<n> 
  {
    cmd &= 0x7F;
    res = send_cmd(CMD55, 0);
    if (res > 1) return res;
  }
  // Select the card and wait for ready 
  deselect();
  if (!select())
  {
    res = 0xFF;
    return res;
  }
  // Send command packet 
  SPI.transfer(0x40 | cmd);  // Start + Command index 
  SPI.transfer((uChar)(arg >> 24));  // Argument[31..24] 
  SPI.transfer((uChar)(arg >> 16));  // Argument[23..16] 
  SPI.transfer((uChar)(arg >> 8));  // Argument[15..8] 
  SPI.transfer((uChar)arg);  // Argument[7..0] 
  n = 0x01;  // Dummy CRC + Stop 
  if (cmd == CMD0) n = 0x95;  // Valid CRC for CMD0(0) + Stop 
  if (cmd == CMD8) n = 0x87;  // Valid CRC for CMD8(0x1AA) Stop 
  SPI.transfer(n);
  // Receive command response 
  if (cmd == CMD12) SPI.transfer(0xFF);  // Skip a stuff byte when stop reading 
  n = 10;  // Wait for a valid response in timeout of 10 attempts 
  do
    res = SPI.transfer(0xFF);
  while ((res & 0x80) && --n);
  return res;  // Return with the response value 
}

/*------------------
 Public Functions
 -------------------*/
void SdSpi::pinSS(uChar pin)
{
  portSS = pin;
}

/*------------------------
 Initialize Disk Drive
 ------------------------*/
DSTATUS SdSpi::disk_initialize ()
{
  uChar n, cmd, ty, ocr[4];
  uInt tmr;

  pinMode(portSS, OUTPUT);
  SPI.begin();
  SPI.setDataMode(SPI_MODE0);
  SPI.setBitOrder(MSBFIRST);
  SPI.setClockDivider(ClockDividerSlow);
  digitalWrite(portSS, LOW);
  for (n = 10; n; n--) SPI.transfer(0xFF);  // 80 dummy clocks 
  ty = 0;
  if (send_cmd(CMD0, 0) == 1)
  {  // Enter Idle state 
    if (send_cmd(CMD8, 0x1AA) == 1)
    {  // SDv2? 
      for (n = 0; n < 4; n++) ocr[n] = SPI.transfer(0xFF);  // Get trailing return value of R7 resp 
      if (ocr[2] == 0x01 && ocr[3] == 0xAA)
      {  // The card can work at vdd range of 2.7-3.6V 
        for (tmr = 1000; tmr; tmr--) 
        {  // Wait for leaving idle state (aCmd41 with HCS bit)
          if (send_cmd(ACMD41, 1UL << 30) == 0)  break;
          delay(wait);
        }
        if (tmr && send_cmd(CMD58, 0) == 0)  // Check CCS bit in the OCR
        {
          for (n = 0; n < 4; n++) ocr[n] = SPI.transfer(0xff);
          ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;  // SDv2 (HC or SC)
        }
      }
    } 
    else {  // SDv1 or MMCv3 
      if (send_cmd(ACMD41, 0) <= 1)
      {
        ty = CT_SD1; 
        cmd = ACMD41;  // SDv1 
      } 
      else
      {
        ty = CT_MMC; 
        cmd = CMD1;  // MMCv3 
      }
      for (tmr = 1000; tmr; tmr--) 
      {  // Wait for leaving idle state
        if (send_cmd(cmd, 0) == 0) break;
        delay(wait);
      }
      if (!tmr || send_cmd(CMD16, 512) != 0) ty = 0;  // Set R/W block length to 512 
    }
  }
  CardType = ty;
  deselect();
  if (ty)  // Initialization succeded 
  {  
    media_status = 0;  // Clear STA_NOINIT 
    SPI.setClockDivider(ClockDividerFast);
  } 
  else media_status = STA_NOINIT;
  return ty > 0 && ty < 10 ? 0 : RES_NOTRDY;
}

/*-----------------
 Get Disk Status
 ------------------*/
DSTATUS SdSpi::disk_status ()
{
  return media_status;
}

/*---------------
 Read Sector(s)
 ----------------*/
DRESULT SdSpi::disk_read (uChar *buff, uLong sector, uInt count)
{
  if (!count) return RES_PARERR;  // if (pdrv || !count) return RES_PARERR;
  if (media_status & STA_NOINIT) return RES_NOTRDY;
  if (!(CardType & CT_BLOCK)) sector *= 512;  // Convert to byte address if needed 
  if (count == 1) // Single block read
  {
    if ((send_cmd(CMD17, sector) == 0) && rcvr_datablock(buff, 512))
    {
      count = 0; // READ_SINGLE_BLOCK 
    }
    else
    {  // Multiple block read 
      if (send_cmd(CMD18, sector) == 0) { // READ_MULTIPLE_BLOCK 
        do
        {
          if (!rcvr_datablock(buff, 512)) break;
          buff += 512;
        } 
        while (--count);
        send_cmd(CMD12, 0);  // STOP_TRANSMISSION 
      }
    }
  }
  deselect();
  return count ? RES_ERROR : RES_OK;
}

/*----------------
 Write Sector(s)
 -----------------*/
DRESULT SdSpi::disk_write (const uChar *buff, uLong sector, uInt count)
{
  if (!count) return RES_PARERR;
  if (media_status & STA_NOINIT) return RES_NOTRDY;
  if (media_status & STA_PROTECT) return RES_WRPRT;
  if (!(CardType & CT_BLOCK)) sector *= 512;  // Convert to byte address if needed 
  if (count == 1)  // Single block write 
  {
    if ((send_cmd(CMD24, sector) == 0) && xmit_datablock(buff, 0xFE)) count = 0; // WRITE_BLOCK 
  }
  else  // Multiple block write 
  {
    if (CardType & CT_SDC) send_cmd(ACMD23, count);
    if (send_cmd(CMD25, sector) == 0) { // WRITE_MULTIPLE_BLOCK 
      do
      {
        if (!xmit_datablock(buff, 0xFC)) break;
        buff += 512;
      } 
      while (--count);
      if (!xmit_datablock(0, 0xFD)) count = 1;  // STOP_TRAN token 
    }
  }
  deselect();
  return count ? RES_ERROR : RES_OK;
}

/*------------------------
 Miscellaneous Functions
 -------------------------*/
/*-------------------------
 stripped in/out control
 -------------------------*/
DRESULT SdSpi::disk_ioctl ()
{
  DRESULT res = RES_ERROR;

  if (media_status & STA_NOINIT) return RES_NOTRDY;
  // Make sure that no pending write process. Do not remove this or written sector might not left updated. 
  if (select())
  {
    deselect();
    res = RES_OK;
  }
  return res;
}

/*-------------------------------------------
 dummy time ** insert your own RTC code **
 -------------------------------------------*/
uLong get_fattime ()
{
  return  ((uLong)(2010 - 1980) << 25)  // Fixed to Jan. 1, 2010 
    | ((uLong)1 << 21)
      | ((uLong)1 << 16)
        | ((uLong)0 << 11)
          | ((uLong)0 << 5)
            | ((uLong)0 >> 1);
}
