
//##### read readme.h #####

#include <Time.h>
#include "TimeDate.h"

TimeDate TD = TimeDate(); 

// ##### helpers #####
unsigned long TimeDate::calcTimeStamp(byte h, byte m, byte s, byte D, byte M, int Y)  // get UNIX time_t Time Stamp from hh:mm:ss DD-MM-YY
{ // UNIX timestamp = number of seconds since 1 Jan 00:00:00 1970 
  // not needed for Arduino; Spark Core Time has no function setTime(h,m,s,D,M,Y))
  // Spark Core using: setTime(calcTimeStamp(h,m,s,D,M,Y))   to set timestamp now()
  if (Y < 100) Y += 2000;
  unsigned long t = 0;
  for (int i = 1970; i < Y; i++) t += yearDay(31, 12, i) * oneDay;
  return t + (yearDay(D, M, Y) - 1) * oneDay + (unsigned long)h * 3600 + (unsigned long)m * 60 + s;
}

boolean TimeDate::leapYear(int y)  // leap year?
{
  int Y = y < 100 ? y + 2000 : y;
  return ((Y % 4 == 0 && Y % 100 != 0) || Y % 400 == 0);
}

int TimeDate::yearDay(unsigned long t)  // year days from timestamp 
{
  return yearDay(day(t), month(t), year(t));
}

int TimeDate::yearDay(byte d, byte m, int y)  // year days from date numbers
{
  int monthdays[13] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334    };
  return monthdays[m - 1] + (monthdays[m - 1] > 31 && leapYear(y) ? 1 : 0) + d;
}

// format a char or byte n (0-99) to char* two digits "00"-"99", or if spatie " 0"-"99"
String TimeDate::twoDigits(byte n, boolean spatie)
{
  return ((n < 10) && spatie ? " " : (String)(n / 10)) + (n % 10);; 
}

String TimeDate::dayName(byte d)  // get the name of the day
{
  return ((String)("SuMoTuWeThFrSa")).substring((d - 1) * 2, d * 2);
}

String TimeDate::monthName(byte m)  // get the name of the month
{
  return ((String)("JanFebMarAprMayJunJulAugSepOctNovDec")).substring((m - 1) * 3, m * 3);
}

String TimeDate::formatTime(unsigned long t, boolean sec, boolean AM)  // format a time char string 
{
  String am = "A";
  byte h = hour(t);
  if (AM)
  {
    if(h > 11)
    {
      h -= 12;
      am = "P";
    }
  }
  String ret = TD.twoDigits(h, false) + ":" + TD.twoDigits(minute(t), false);
  if (sec)
  {
    ret += ":" + TD.twoDigits(second(t), false);
  }
  if (AM) ret += " " + am + "M";
  return ret;
}

String TimeDate::formatDate(unsigned long t, String format)  //format a date char string 
{
  String ret = format;
  ret.replace("xd", TD.twoDigits(day(t), true));  // day as 1-31
  ret.replace("dd", TD.twoDigits(day(t), false));  // day as 01-31
  ret.replace("xm", TD.twoDigits(month(t), true));  // month as 1-12
  ret.replace("mm", TD.twoDigits(month(t), false));  // month as 01-12
  ret.replace("yyyy", (String)year(t));  // year as 1970-2099
  ret.replace("DD", TD.dayName(weekday(t)));  // day name Su-Za
  ret.replace("MMM", TD.monthName(month(t)));  // month name Jan-Dec
  return ret;
}

// ##### DST #####
void TimeDate::initDst(byte s1, byte s2, byte s3, byte s4, byte s5,  // ...Month, ...MonthDay, ...Week, ...WeekDay, ...Hour
                       byte e1, byte e2, byte e3, byte e4, byte e5,  // s = DST start, e = DST end
                       long utc, int dst) // time zone and DST in seconds
{
  startMonth = s1; 
  startMonthDay = s2; 
  startWeek = s3; 
  startWeekDay = s4; 
  startHour = s5;
  endMonth = e1; 
  endMonthDay = e2; 
  endWeek = e3; 
  endWeekDay = e4; 
  endHour = e5;
  utcOffset = utc;
  dstOffset = dst;
  // Globals for DST/Summer Time 
  startDst = 0; 
  endDst = 0;
}

long TimeDate::localTimeOffset(unsigned long utcTime)  // determine time offset: time zone or time zone + DST offset
{  // isDst, startDst and endDst are also calculated
  return TD.calcDst(utcTime) ? utcOffset + dstOffset : utcOffset;
}

// PRIVATE calculate the globals 'startDst', 'endDst' and (re)set the flag 'isDst', if DST then return true
boolean TimeDate::calcDst(unsigned long utcTime)  // use 'localTimeOffset' to set the globals 
{
  // Globals Time Stamp dst.startDst and dst.endDst, if 0 then not initialized or no DST or Summertime
  startDst = 0; 
  endDst = 0;
  isDst = false;
  if ((startMonthDay == 0 && startWeekDay == 0) || (endMonthDay == 0 && endWeekDay == 0)) return false;  // no DST or Summertime
  if (startMonth == 0 || startWeekDay == 0 || endMonth == 0 || endWeekDay == 0) return false;  // no DST or Summertime
  // calculate time stamp start DST and  end DST
  startDst = findDstDay(year(utcTime), startMonth, startMonthDay, startWeek, startWeekDay, startHour); 
  endDst = findDstDay(year(utcTime), endMonth, endMonthDay, endWeek, endWeekDay, endHour);
  if (TimeDate::startDst < TimeDate::endDst)  // is't summer or winter in the middle of the year
  {  // summer
    if ((utcTime + utcOffset) >= TimeDate::startDst && (utcTime + utcOffset + dstOffset) < TimeDate::endDst) isDst = true; 
  }
  else
  {  // winter (see example "Lord Howe Island" DST)
    if (!((utcTime + utcOffset) < TimeDate::startDst && (utcTime + utcOffset + dstOffset) >= TimeDate::endDst)) isDst = true;
  }
  return isDst;
  // DST is calculated and stored in Globals isDst, startDst and endDst
}

// PRIVATE find DST start or end date
unsigned long TimeDate::findDstDay(int _year, byte _month, byte monthDay, byte week, byte weekDay, byte _hour)
{
  unsigned long dstDate = TimeDate::calcTimeStamp(0, 0, 0, 1, 1, _year);  // calc time stamp 1 Jan this year 00:00:00
  if (monthDay)
  {  // fixed date DST
    dstDate += TD.yearDay(monthDay - 1, _month, _year) * oneDay;  //  date in seconds
  }
  else
  {  // calcualated DST
    if (week == 0)
    {  // day in last week of the month
      dstDate += (TD.yearDay(1, _month + 1, _year) - 1) * oneDay;  // ad year days to the first day of next month in seconds
      while (weekday(dstDate) != weekDay) dstDate -= oneDay; // now it's right weekday of the last week
      if (month(dstDate) != _month) dstDate -= oneWeek;
    }   
    else
    {  // day in week 1, 2, 3 or 4 of the month
      dstDate += (TD.yearDay(1, _month, _year) - 1) * oneDay;  // year days to the first day of month in seconds
      while (weekday(dstDate) != weekDay) dstDate += oneDay;  // now it's right weekday of the first week 
      dstDate += (week - 1) * oneWeek;  // now it's right weekday of the right week
    }
  }
  return dstDate += _hour * 3600;  // adding the time seconds after midnight
}

// ##### SUN #####
void TimeDate::initSun(float lt, float ln, int utc)  // init Sun with location and time zone in seconds
{
  lat = lt; 
  lon = ln;
  timeZone = utc / 3600;
}

void TimeDate::CalcSunTimes(byte d, byte m, int y, unsigned int dstoffset)  // calculate SunRise and SunSet in local time 
{  
  unsigned long i = - 7 * (y + (m + 9) / 12) / 4 + 275 * m / 9 + d;
  i += (unsigned long)y * 367;
  float yd = (float)i - 730531.5 + 12 / 24.0;
  float L = rangeAngle(280.461 * DEG_TO_RAD + .9856474 * DEG_TO_RAD * yd);
  float M = rangeAngle(357.528 * DEG_TO_RAD + .9856003 * DEG_TO_RAD * yd);
  float lambda = rangeAngle(L + 1.915 * DEG_TO_RAD * sin(M) + .02 * DEG_TO_RAD * sin(2 * M));
  float obliq = 23.439 * DEG_TO_RAD - .0000004 * DEG_TO_RAD * yd;
  float alpha = atan2(cos(obliq) * sin(lambda), cos(lambda));
  float delta = asin(sin(obliq) * sin(lambda));
  float LL = L - alpha;
  if (L < PI) LL += 2.0 * PI;
  float equation = 1440.0 * (1.0 - LL / PI / 2.0);
  float dfo = DEG_TO_RAD*(0.5 * 0.53 + 34.0 / 60.0); 
  if (lat < 0.0) dfo = -dfo;
  float fo = tan(delta + dfo) * tan(lat * DEG_TO_RAD);
  if (fo>0.99999) fo = 1.0;
  float ha = asin(fo) + PI / 2.0;
  float daylen = RAD_TO_DEG * ha / 7.5;
  if (daylen < 0.0001) daylen = 0.0;
  sunRiseDec = 12.0 - 12.0 * ha / PI + timeZone - lon / 15.0 + equation / 60.0;
  sunSetDec = 12.0 + 12.0 * ha / PI + timeZone - lon / 15.0 + equation / 60.0;
  sunRiseDec += dstoffset / 3600;
  sunSetDec += dstoffset / 3600;
  if (sunRiseDec > 24.0) sunRiseDec -= 24.0;
  if (sunSetDec > 24.0) sunSetDec -= 24.0;
  sunRiseHour = sunRiseDec;
  sunRiseMinute = (sunRiseDec - sunRiseHour) * 60;
  sunSetHour = sunSetDec;
  sunSetMinute = (sunSetDec - sunSetHour) * 60;
}

float TimeDate::rangeAngle(float x)  // PRIVATE for CalcSunTimes()
{
  float b = 0.5 * x / PI;
  float a = 2.0 * PI * (b - (long)(b));
  if (a < 0) a = 2.0 * PI + a;
  return a;
}

